import Papa from 'papaparse';
import {
    isStatePopulated,
    isStateClear,
    PartialStateUpdate,
    ExporterState,
    isStateAbleToRequest
} from '../state';
import { Status } from '../state';
import { LevelState } from '../state/level';
import * as landsbankinn from '../landsbankinn';
import {
    fetchAccounts,
    downloadTransactions,
    OutputFormat,
    RawTransaction
} from '../landsbankinn';
import {
    Actions,
    DownloadTransactionReply,
    DownloadTransactionsMessage,
    EnsureStateMessage,
    EnsureStateReply,
    GetAccountsReply
} from './messages';

type GroupedRawTransaction = RawTransaction & { txnAccountNumber: string };
type EnsureStateAction = Required<Pick<EnsureStateMessage, 'clientState'>>;
type DownloadTransactionsAction = Omit<DownloadTransactionsMessage, 'action'>;
type TransactionsByAccount = Record<string, RawTransaction[] | undefined>;

// State singleton (move to its own file if we eventually get more
// than 1 background script)
const State = new LevelState();

async function ensureDataPopulated(state: ExporterState) {
    if (isStateAbleToRequest(state) && !isStatePopulated(state)) {
        const accounts = await fetchAccounts(state);
        await State.update({ status: Status.Populated, accounts });
        console.info('Populated account data');
    }
}

async function ensureAuthState(message: EnsureStateAction) {
    const clientState: PartialStateUpdate = message.clientState!;
    let state = await State.current;

    if (isStateClear(state)) {
        await State.update({ status: Status.Updating });
        const session = await landsbankinn.authSession();
        await State.update({
            status: Status.AbleToRequest,
            authToken: session.accessToken,
            ...clientState
        });

        console.info('Updated auth state from client and auth session');
    }

    return await State.current;
}

async function ensureState(message: EnsureStateAction): Promise<EnsureStateReply> {
    console.info('Ensuring state');
    const state = await ensureAuthState(message);
    await ensureDataPopulated(state);
    return { state: 'Populated' };
}

async function getAccounts(): Promise<GetAccountsReply> {
    const state = await State.current;
    let message: any;

    await ensureDataPopulated(state);

    // Should always be populated, but we can be careful anyway.
    if (isStatePopulated(state)) {
        message = {
            accounts: (await State.current).accounts,
            error: null
        };
    } else {
        message = {
            accounts: undefined,
            error: new Error('state not yet populated')
        };
    }

    return message;
}

interface DownloadResult {
    accountNumber: string;
    transactions: landsbankinn.RawTransaction[];
}

const getAllTransactions = async (message: DownloadTransactionsAction) =>
    await Promise.all(
        message.accounts.map(async (accountNumber) => {
            const transactions = (
                await downloadTransactions({
                    accountNumber: accountNumber,
                    dateFrom: message.dateFrom,
                    dateTo: message.dateTo,
                    outputFormat: OutputFormat.CSV
                })
            ).map((txn) => ({ txnAccountNumber: accountNumber, ...txn }));

            return transactions;
        })
    );

const mapToCsv = (transactions: GroupedRawTransaction[]) => Papa.unparse(transactions);

async function downloadStatements(
    message: DownloadTransactionsMessage
): Promise<DownloadTransactionReply> {
    console.info('Initiating download for', message.accounts);
    const state = await ensureAuthState({ clientState: await State.current });
    await ensureDataPopulated(state);

    // Should always be populated, but we can be careful anyway.
    if (isStatePopulated(state)) {
        const transactions = (await getAllTransactions(message)).flat();
        return { csv: mapToCsv(transactions), error: null };
    } else {
        return {
            csv: undefined,
            error: new Error('state not yet populated')
        };
    }
}

browser.runtime.onMessage.addListener(async (message) => {
    console.log('Received message:', message);
    switch (message.action) {
        case Actions.EnsureState:
            return await ensureState(message as EnsureStateMessage);
        case Actions.GetAccounts:
            return await getAccounts();
        case Actions.DownloadTransactions:
            return await downloadStatements(message);
    }
});

export {};
